mod game;

use anyhow::{bail, Result};
use clap::Parser;
use galactica_content::Content;
use galactica_playeragent::PlayerAgent;
use galactica_render::{InputEvent, RenderInput};
use galactica_system::phys::PhysImage;
use galactica_util::constants::ASSET_CACHE;
use log::LevelFilter;
use log4rs::{
	append::console::ConsoleAppender,
	config::{Appender, Logger, Root},
	encode::pattern::PatternEncoder,
	Config,
};
use std::{
	fs,
	path::{Path, PathBuf},
	sync::Arc,
};
use winit::{
	event::{ElementState, Event, KeyboardInput, MouseButton, MouseScrollDelta, WindowEvent},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
	/// Run the asset packer
	#[arg(long, action, group = "run")]
	pack: bool,

	/// Check content in the default directory
	#[arg(long, action, group = "run")]
	check_content: bool,
}

mod cli;

fn set_up_logger(_args: &Args) -> log4rs::Handle {
	// h: highlight with level color
	// :<0.128 right align, min 0, max 128
	let stdout = ConsoleAppender::builder()
		.encoder(Box::new(PatternEncoder::new(
			"{d(%H:%M:%S)} | {({h({l})}):5.5} | \x1b[0;34m{M}\x1b[0m — {m:<0.128}{n}",
		)))
		.build();

	let config = Config::builder()
		.appender(Appender::builder().build("stdout", Box::new(stdout)))
		.logger(Logger::builder().build("wgpu_core", LevelFilter::Warn))
		.logger(Logger::builder().build("wgpu_hal", LevelFilter::Warn))
		.logger(Logger::builder().build("naga", LevelFilter::Warn))
		.logger(Logger::builder().build("cosmic_text", LevelFilter::Warn))
		.logger(Logger::builder().build("winit", LevelFilter::Warn))
		.logger(Logger::builder().build("mio", LevelFilter::Warn))
		.build(Root::builder().appender("stdout").build(LevelFilter::Trace))
		.unwrap();

	log4rs::init_config(config).unwrap()
}

fn main() -> Result<()> {
	let args = Args::parse();

	let _ = set_up_logger(&args);

	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	if args.pack {
		cli::packer::run(Path::new("./assets/render"), cache_dir)?;
	} else if args.check_content {
		let atlas_index = cache_dir.join("spriteatlas.toml");
		let _ = Content::load_dir(
			PathBuf::from("./content"),
			PathBuf::from("./assets"),
			atlas_index.clone(),
		)?;
		println!("Content was loaded without errors.")
	} else {
		// Run game
		if let Err(err) = try_main() {
			eprintln!("Galactica failed with an error");

			let mut i = 0;
			for e in err.chain().rev() {
				eprintln!("{i:02}: {}", e);
				i += 1;
			}
			std::process::exit(1);
		}
	}

	return Ok(());
}

fn try_main() -> Result<()> {
	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	let atlas_index = cache_dir.join("spriteatlas.toml");

	if !atlas_index.exists() {
		bail!("Could not find sprite atlas!");
	}

	// TODO: pretty error if missing (also in cli)
	let content = Arc::new(Content::load_dir(
		PathBuf::from("./content"),
		PathBuf::from("./assets"),
		atlas_index,
	)?);

	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();
	let mut gpu = pollster::block_on(galactica_render::GPUState::new(window, content.clone()))?;
	gpu.init(&content);

	// TODO: don't clone content
	let mut game = game::Game::new(content.clone());
	let p = game.make_player();

	let player = Arc::new(PlayerAgent::new(&content, p.0));
	let mut phys_img = Arc::new(PhysImage::new());

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				match gpu.render(RenderInput {
					current_time: game.get_current_time(),
					ct: content.clone(),
					phys_img: phys_img.clone(),
					player: player.clone(),
					// TODO: this is a hack for testing.
					current_system: content.systems.values().next().unwrap().clone(),
					timing: game.get_timing().clone(),
				}) {
					Ok(_) => {}
					Err(wgpu::SurfaceError::Lost) => gpu.resize(&content),
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				game.step(&phys_img);
				game.update_image(Arc::get_mut(&mut phys_img).unwrap());
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window().id() => match event {
				WindowEvent::Focused(_state) => {
					// TODO: handle focus loss
				}
				WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
				WindowEvent::KeyboardInput {
					input:
						KeyboardInput {
							state,
							virtual_keycode: Some(key),
							..
						},
					..
				} => {
					let directive = gpu
						.process_input(
							RenderInput {
								current_time: game.get_current_time(),
								ct: content.clone(),
								phys_img: phys_img.clone(),
								player: player.clone(),
								current_system: content.systems.values().next().unwrap().clone(),
								timing: game.get_timing().clone(),
							},
							InputEvent::Keyboard {
								down: state == &ElementState::Pressed,
								key: *key,
							},
						)
						.unwrap();
					game.apply_directive(directive, &player);
				}
				WindowEvent::CursorMoved { position, .. } => {
					let directive = gpu
						.process_input(
							RenderInput {
								current_time: game.get_current_time(),
								ct: content.clone(),
								phys_img: phys_img.clone(),
								player: player.clone(),
								current_system: content.systems.values().next().unwrap().clone(),
								timing: game.get_timing().clone(),
							},
							InputEvent::MouseMove(position.cast()),
						)
						.unwrap();
					game.apply_directive(directive, &player);
				}
				WindowEvent::MouseInput { state, button, .. } => {
					let down = state == &ElementState::Pressed;
					let event = match button {
						MouseButton::Left => Some(InputEvent::MouseLeftClick(down)),
						MouseButton::Right => Some(InputEvent::MouseRightClick(down)),
						_ => None,
					};
					if let Some(event) = event {
						let directive = gpu
							.process_input(
								RenderInput {
									current_time: game.get_current_time(),
									ct: content.clone(),
									phys_img: phys_img.clone(),
									player: player.clone(),
									current_system: content
										.systems
										.values()
										.next()
										.unwrap()
										.clone(),
									timing: game.get_timing().clone(),
								},
								event,
							)
							.unwrap();
						game.apply_directive(directive, &player);
					}
				}
				WindowEvent::MouseWheel { delta, .. } => {
					let directive = gpu
						.process_input(
							RenderInput {
								current_time: game.get_current_time(),
								ct: content.clone(),
								phys_img: phys_img.clone(),
								player: player.clone(),
								current_system: content.systems.values().next().unwrap().clone(),
								timing: game.get_timing().clone(),
							},
							InputEvent::Scroll(match delta {
								MouseScrollDelta::LineDelta(_h, v) => *v,
								MouseScrollDelta::PixelDelta(v) => v.x as f32,
							}),
						)
						.unwrap();
					game.apply_directive(directive, &player);
				}
				WindowEvent::Resized(_) => {
					gpu.resize(&content);
				}
				WindowEvent::ScaleFactorChanged { .. } => {
					gpu.resize(&content);
					gpu.window().request_redraw();
				}
				_ => {}
			},
			_ => {}
		}
	});
}
