use log::error;
use rhai::{FnNamespace, FuncRegistration, ImmutableString, Module};
use std::{cell::RefCell, rc::Rc};

use super::super::Rect;
use crate::ui::{elements::UiScrollbox, UiElement, UiState};

pub fn build_scrollbox_module(state_src: Rc<RefCell<UiState>>) -> Module {
	let mut module = Module::new();
	module.set_id("GalacticaScrollboxModule");

	let state = state_src.clone();
	let _ = FuncRegistration::new("add")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString, rect: Rect| {
			let mut ui_state = state.borrow_mut();

			if ui_state.elements.contains_key(&name) {
				error!("tried to make a scrollbox using an existing name `{name}`");
				return;
			}

			ui_state.add_element(UiElement::Scrollbox(UiScrollbox::new(name.clone(), rect)));
		});

	let state = state_src.clone();
	let _ = FuncRegistration::new("add_element")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(
			&mut module,
			move |name: ImmutableString, target: ImmutableString| {
				let mut ui_state = state.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Scrollbox(_)) => {
						match ui_state.get_mut_by_name(&target) {
							Some(UiElement::Text(_)) | Some(UiElement::Sprite(_)) => {
								let e = match ui_state.remove_element_incomplete(&target) {
									Some(UiElement::Sprite(s)) => {
										Rc::new(RefCell::new(UiElement::Sprite(s)))
									}
									Some(UiElement::Text(t)) => {
										Rc::new(RefCell::new(UiElement::Text(t)))
									}
									_ => unreachable!(),
								};

								// Add a subelement pointing to this sprite
								ui_state.add_element(UiElement::SubElement {
									parent: name.clone(),
									element: e.clone(),
								});

								// Add this sprite to a scrollbox
								match ui_state.get_mut_by_name(&name) {
									Some(UiElement::Scrollbox(s)) => {
										s.add_element(e);
									}
									_ => unreachable!(),
								};
							}
							Some(_) => {
								error!("cannot add `{name}` to scrollbox `{name}`, invalid type.")
							}
							None => {
								error!("called `scrollbox::add_element` with a non-existing target `{target}`")
							}
						}
					}
					_ => {
						error!("called `scrollbox::add_element` on an invalid name `{name}`")
					}
				}
			},
		);

	return module;
}
