use nalgebra::{Point2, Vector2};
use rhai::{CustomType, TypeBuilder};
use winit::{
	dpi::{LogicalSize, PhysicalPosition},
	window::Window,
};

use super::{anchor::Anchor, vector::UiVector};
use crate::RenderState;

#[derive(Debug, Clone)]
pub struct Rect {
	pub pos: Point2<f32>,
	pub dim: Vector2<f32>,
	pub anchor_self: Anchor,
	pub anchor_parent: Anchor,
}

impl Rect {
	pub fn new(x: f32, y: f32, w: f32, h: f32, anchor_self: Anchor, anchor_parent: Anchor) -> Self {
		Self {
			pos: Point2::new(x, y),
			dim: Vector2::new(w, h),
			anchor_self,
			anchor_parent,
		}
	}

	pub fn to_centered(&self, window: &Window, ui_scale: f32) -> CenteredRect {
		let w: LogicalSize<f32> = window.inner_size().to_logical(window.scale_factor());
		let w = Vector2::new(w.width, w.height);

		let mut pos = self.pos * ui_scale;
		let dim = self.dim * ui_scale;

		// Origin
		match self.anchor_parent {
			Anchor::Center => {}

			Anchor::NorthWest => {
				pos += Vector2::new(-w.x, w.y) / 2.0;
			}

			Anchor::SouthWest => {
				pos += Vector2::new(-w.x, -w.y) / 2.0;
			}

			Anchor::NorthEast => {
				pos += Vector2::new(w.x, w.y) / 2.0;
			}

			Anchor::SouthEast => {
				pos += Vector2::new(w.x, -w.y) / 2.0;
			}
		}

		// Offset for self dimensions
		match self.anchor_self {
			Anchor::Center => {}

			Anchor::NorthWest => {
				pos += Vector2::new(dim.x, -dim.y) / 2.0;
			}

			Anchor::NorthEast => {
				pos += Vector2::new(-dim.x, -dim.y) / 2.0;
			}

			Anchor::SouthWest => {
				pos += Vector2::new(dim.x, dim.y) / 2.0;
			}

			Anchor::SouthEast => {
				pos += Vector2::new(-dim.x, dim.y) / 2.0;
			}
		};

		return CenteredRect { pos, dim };
	}
}

impl CustomType for Rect {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("Rect")
			.with_fn("Rect", Self::new)
			.with_fn("pos", |s: &mut Self| UiVector::new(s.pos.x, s.pos.y))
			.with_fn("dim", |s: &mut Self| UiVector::new(s.dim.x, s.dim.y));
	}
}

/// Represents a rectangular region, in absolute coordinates relative to the screen center.
#[derive(Debug, Clone, Copy)]
pub(crate) struct CenteredRect {
	/// The position of the center of this rectangle, in post-scale logical pixels,
	/// relative to (0, 0) in the center of the screen.
	pub pos: Point2<f32>,

	/// The width and height of this rectangle, in post-scale logical pixels.
	pub dim: Vector2<f32>,
}

impl CenteredRect {
	pub fn contains_point(&self, pt: Point2<f32>) -> bool {
		let ne = self.pos + Vector2::new(-self.dim.x, self.dim.y) / 2.0;
		let sw = self.pos + Vector2::new(self.dim.x, -self.dim.y) / 2.0;
		return (pt.y < ne.y && pt.y > sw.y) && (pt.x > ne.x && pt.x < sw.x);
	}

	pub fn contains_mouse(&self, state: &RenderState, mouse_pos: &PhysicalPosition<f32>) -> bool {
		let fac = state.window.scale_factor() as f32;
		let window_size = Vector2::new(
			state.window_size.width as f32 / fac,
			state.window_size.height as f32 / fac,
		);

		let mouse_pos = Point2::new(
			mouse_pos.x / fac - window_size.x / 2.0,
			window_size.y / 2.0 - mouse_pos.y / fac,
		);

		return self.contains_point(mouse_pos);
	}
}
