use std::sync::Arc;

use galactica_content::{AnimationState, Faction, Projectile, SpriteAutomaton};
use rand::Rng;
use rapier2d::{dynamics::RigidBodyHandle, geometry::ColliderHandle};

use crate::phys::{physsim::NewObjects, PhysStepResources, PhysWrapper};

use super::PhysEffect;

/// A single projectile in this simulation
#[derive(Debug, Clone)]
pub struct PhysProjectile {
	/// This projectile's game data
	pub content: Arc<Projectile>,

	/// This projectile's sprite animation state
	anim: SpriteAutomaton,

	/// The remaining lifetime of this projectile, in seconds
	lifetime: f32,

	/// The faction this projectile belongs to
	pub faction: Arc<Faction>,

	/// This projectile's rigidbody
	pub rigid_body: RigidBodyHandle,

	/// This projectile's collider
	pub collider: ColliderHandle,

	/// This projectile's size variation
	pub size_rng: f32,

	/// If true, this projectile has been destroyed
	/// and is waiting to be deallocated.
	is_destroyed: bool,
}

impl PhysProjectile {
	/// Create a new projectile
	pub fn new(
		content: Arc<Projectile>,
		rigid_body: RigidBodyHandle,
		faction: Arc<Faction>,
		collider: ColliderHandle,
	) -> Self {
		let mut rng = rand::thread_rng();
		let size_rng = content.size_rng;
		let lifetime = content.lifetime;
		PhysProjectile {
			anim: SpriteAutomaton::new(content.sprite.clone()),
			rigid_body,
			collider,
			content,
			lifetime,
			faction,
			size_rng: rng.gen_range(-size_rng..=size_rng),
			is_destroyed: false,
		}
	}

	/// Process this projectile's state after `t` seconds
	pub(in crate::phys) fn step(
		&mut self,
		res: &PhysStepResources,
		new: &mut NewObjects,
		wrapper: &mut PhysWrapper,
	) {
		self.lifetime -= res.t;
		self.anim.step(res.t);

		if self.lifetime <= 0.0 {
			self.destroy(new, wrapper, true);
		}
	}

	/// Destroy this projectile without creating an expire effect
	pub(in crate::phys) fn destroy_silent(
		&mut self,
		new: &mut NewObjects,
		wrapper: &mut PhysWrapper,
	) {
		self.destroy(new, wrapper, false);
	}

	/// Destroy this projectile
	fn destroy(&mut self, new: &mut NewObjects, wrapper: &mut PhysWrapper, expire: bool) {
		if self.is_destroyed {
			return;
		}

		let rb = wrapper.get_rigid_body(self.rigid_body).unwrap();
		if expire {
			match &self.content.expire_effect {
				None => {}
				Some(effect) => {
					new.effects.push(PhysEffect::new(
						wrapper,
						effect.clone(),
						*rb.translation(),
						self.rigid_body,
						None,
					));
				}
			}
		};

		self.is_destroyed = true;
		wrapper.remove_rigid_body(self.rigid_body);
	}

	/// Should this effect be deleted?
	pub(in crate::phys) fn should_remove(&self) -> bool {
		self.is_destroyed
	}
}

impl PhysProjectile {
	/// Get this projectile's animation state
	pub fn get_anim_state(&self) -> AnimationState {
		self.anim.get_texture_idx()
	}
}
